@@ -30,6 +30,7 @@
 #include <netinet/in.h>
 #include <netdb.h>
 #include "netdb-compat.h"
+#include <inttypes.h>
 #include <stdio.h>
 #include <fcntl.h>
 #include <syslog.h>
@@ -44,15 +45,143 @@
 #include <time.h>
 
 #include <linux/ioctl.h>
+
+#if HAVE_NETLINK
+#include "nbd-netlink.h"
+#include <netlink/netlink.h>
+#include <netlink/genl/genl.h>
+#include <netlink/genl/ctrl.h>
+#endif
+
 #define MY_NAME "nbd_client"
 #include "cliserv.h"
 
+#if HAVE_GNUTLS && !defined(NOTLS)
+#include "crypto-gnutls.h"
+#endif
+
 #ifdef WITH_SDP
 #include <sdp_inet.h>
 #endif
 
 #define NBDC_DO_LIST 1
 
+#if HAVE_NETLINK
+static int callback(struct nl_msg *msg, void *arg) {
+	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));
+	struct nlattr *msg_attr[NBD_ATTR_MAX + 1];
+	int ret;
+	uint32_t index;
+
+	ret = nla_parse(msg_attr, NBD_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
+			genlmsg_attrlen(gnlh, 0), NULL);
+	if (ret)
+		err("Invalid response from the kernel\n");
+	if (!msg_attr[NBD_ATTR_INDEX])
+		err("Did not receive index from the kernel\n");
+	index = nla_get_u32(msg_attr[NBD_ATTR_INDEX]);
+	printf("Connected /dev/nbd%d\n", (int)index);
+	return NL_OK;
+}
+
+static struct nl_sock *get_nbd_socket(int *driver_id) {
+	struct nl_sock *socket;
+
+	socket = nl_socket_alloc();
+	if (!socket)
+		err("Couldn't allocate netlink socket\n");
+
+	if (genl_connect(socket))
+		err("Couldn't connect to the generic netlink socket\n");
+	*driver_id = genl_ctrl_resolve(socket, "nbd");
+	if (*driver_id < 0)
+		err("Couldn't resolve the nbd netlink family, make sure the nbd module is loaded and your nbd driver supports the netlink interface.\n");
+	return socket;
+}
+
+static void netlink_configure(int index, int *sockfds, int num_connects,
+			      u64 size64, int blocksize, uint16_t flags,
+			      int timeout) {
+	struct nl_sock *socket;
+	struct nlattr *sock_attr;
+	struct nl_msg *msg;
+	int driver_id, i;
+
+	socket = get_nbd_socket(&driver_id);
+	nl_socket_modify_cb(socket, NL_CB_VALID, NL_CB_CUSTOM, callback, NULL);
+
+	msg = nlmsg_alloc();
+	if (!msg)
+		err("Couldn't allocate netlink message\n");
+	genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, driver_id, 0, 0,
+		    NBD_CMD_CONNECT, 0);
+	if (index >= 0)
+		NLA_PUT_U32(msg, NBD_ATTR_INDEX, index);
+	NLA_PUT_U64(msg, NBD_ATTR_SIZE_BYTES, size64);
+	NLA_PUT_U64(msg, NBD_ATTR_BLOCK_SIZE_BYTES, blocksize);
+	NLA_PUT_U64(msg, NBD_ATTR_SERVER_FLAGS, flags);
+	NLA_PUT_U64(msg, NBD_ATTR_TIMEOUT, timeout);
+
+	sock_attr = nla_nest_start(msg, NBD_ATTR_SOCKETS);
+	if (!sock_attr)
+		err("Couldn't nest the sockets for our connection\n");
+	for (i = 0; i < num_connects; i++) {
+		struct nlattr *sock_opt;
+		sock_opt = nla_nest_start(msg, NBD_SOCK_ITEM);
+		if (!sock_opt)
+			err("Couldn't nest the sockets for our connection\n");
+		NLA_PUT_U32(msg, NBD_SOCK_FD, sockfds[i]);
+		nla_nest_end(msg, sock_opt);
+	}
+	nla_nest_end(msg, sock_attr);
+
+	if (nl_send_sync(socket, msg) < 0)
+		err("Failed to setup device, check dmesg\n");
+	return;
+nla_put_failure:
+	err("Failed to create netlink message\n");
+}
+
+static void netlink_disconnect(char *nbddev) {
+	struct nl_sock *socket;
+	struct nl_msg *msg;
+	int driver_id;
+
+	int index = -1;
+	if (nbddev) {
+		if (sscanf(nbddev, "/dev/nbd%d", &index) != 1)
+			err("Invalid nbd device target\n");
+	}
+	if (index < 0)
+		err("Invalid nbd device target\n");
+
+	socket = get_nbd_socket(&driver_id);
+
+	msg = nlmsg_alloc();
+	if (!msg)
+		err("Couldn't allocate netlink message\n");
+	genlmsg_put(msg, NL_AUTO_PORT, NL_AUTO_SEQ, driver_id, 0, 0,
+		    NBD_CMD_DISCONNECT, 0);
+	NLA_PUT_U32(msg, NBD_ATTR_INDEX, index);
+	if (nl_send_sync(socket, msg) < 0)
+		err("Failed to disconnect device, check dmsg\n");
+	nl_socket_free(socket);
+	return;
+nla_put_failure:
+	err("Failed to create netlink message\n");
+}
+#else
+static void netlink_configure(int index, int *sockfds, int num_connects,
+			      u64 size64, int blocksize, uint16_t flags,
+			      int timeout)
+{
+}
+
+static void netlink_disconnect(char *nbddev)
+{
+}
+#endif /* HAVE_NETLINK */
+
 int check_conn(char* devname, int do_print) {
 	char buf[256];
 	char* p;
@@ -168,30 +297,76 @@ int openunix(const char *path) {
 	return sock;
 }
 
-void ask_list(int sock) {
+void send_request(int sock, uint32_t opt, ssize_t datasize, void* data) {
+	struct {
+		uint64_t magic;
+		uint32_t opt;
+		uint32_t datasize;
+	} __attribute__((packed)) header = {
+		ntohll(opts_magic),
+		ntohl(opt),
+		ntohl(datasize),
+	};
+	if(datasize < 0) {
+		datasize = strlen((char*)data);
+		header.datasize = htonl(datasize);
+	}
+	writeit(sock, &header, sizeof(header));
+	if(data != NULL) {
+		writeit(sock, data, datasize);
+	}
+}
+
+void send_info_request(int sock, uint32_t opt, int n_reqs, uint16_t* reqs, char* name) {
+	uint16_t rlen = htons(n_reqs);
+	uint32_t nlen = htonl(strlen(name));
+
+	send_request(sock, opt, sizeof(uint32_t) + strlen(name) + sizeof(uint16_t) + n_reqs * sizeof(uint16_t), NULL);
+	writeit(sock, &nlen, sizeof(nlen));
+	writeit(sock, name, strlen(name));
+	writeit(sock, &rlen, sizeof(rlen));
+	if(n_reqs > 0) {
+		writeit(sock, reqs, n_reqs * sizeof(uint16_t));
+	}
+}
+
+struct reply {
+	uint64_t magic;
 	uint32_t opt;
+	uint32_t reply_type;
+	uint32_t datasize;
+	char data[];
+} __attribute__((packed));
+
+struct reply* read_reply(int sock) {
+	struct reply *retval = malloc(sizeof(struct reply));
+	readit(sock, retval, sizeof(*retval));
+	retval->magic = ntohll(retval->magic);
+	retval->opt = ntohl(retval->opt);
+	retval->reply_type = ntohl(retval->reply_type);
+	retval->datasize = ntohl(retval->datasize);
+	if (retval->magic != rep_magic) {
+		fprintf(stderr, "E: received invalid negotiation magic %" PRIu64 " (expected %" PRIu64 ")", retval->magic, rep_magic);
+		exit(EXIT_FAILURE);
+	}
+	if (retval->datasize > 0) {
+		retval = realloc(retval, sizeof(struct reply) + retval->datasize);
+		readit(sock, &(retval->data), retval->datasize);
+	}
+	return retval;
+}
+
+void ask_list(int sock) {
 	uint32_t opt_server;
 	uint32_t len;
+	uint32_t lenn;
 	uint32_t reptype;
 	uint64_t magic;
 	int rlen;
 	const int BUF_SIZE = 1024;
 	char buf[BUF_SIZE];
 
-	magic = ntohll(opts_magic);
-	if (write(sock, &magic, sizeof(magic)) < 0)
-		err("Failed/2.2: %m");
-
-	/* Ask for the list */
-	opt = htonl(NBD_OPT_LIST);
-	if(write(sock, &opt, sizeof(opt)) < 0) {
-		err("writing list option failed: %m");
-	}
-	/* Send the length (zero) */
-	len = htonl(0);
-	if(write(sock, &len, sizeof(len)) < 0) {
-		err("writing length failed: %m");
-	}
+	send_request(sock, NBD_OPT_LIST, 0, NULL);
 	/* newline, move away from the "Negotiation:" line */
 	printf("\n");
 	do {
@@ -233,46 +408,65 @@ void ask_list(int sock) {
 			}
 			exit(EXIT_FAILURE);
 		} else {
-			if(len) {
+			if(reptype != NBD_REP_ACK) {
 				if(reptype != NBD_REP_SERVER) {
 					err("Server sent us a reply we don't understand!");
 				}
-				if(read(sock, &len, sizeof(len)) < 0) {
+				if(read(sock, &lenn, sizeof(lenn)) < 0) {
 					fprintf(stderr, "\nE: could not read export name length from server\n");
 					exit(EXIT_FAILURE);
 				}
-				len=ntohl(len);
-				if (len >= BUF_SIZE) {
+				lenn=ntohl(lenn);
+				if (lenn >= BUF_SIZE) {
 					fprintf(stderr, "\nE: export name on server too long\n");
 					exit(EXIT_FAILURE);
 				}
-				if(read(sock, buf, len) < 0) {
+				if(read(sock, buf, lenn) < 0) {
 					fprintf(stderr, "\nE: could not read export name from server\n");
 					exit(EXIT_FAILURE);
 				}
-				buf[len] = 0;
-				printf("%s\n", buf);
+				buf[lenn] = 0;
+				printf("%s", buf);
+				len -= lenn;
+				len -= sizeof(lenn);
+				if(len > 0) {
+					if(read(sock, buf, len) < 0) {
+						fprintf(stderr, "\nE: could not read export description from server\n");
+						exit(EXIT_FAILURE);
+					}
+					buf[len] = 0;
+					printf(": %s\n", buf);
+				} else {
+					printf("\n");
+				}
 			}
 		}
 	} while(reptype != NBD_REP_ACK);
-	opt=htonl(NBD_OPT_ABORT);
-	len=htonl(0);
-	magic=htonll(opts_magic);
-	if (write(sock, &magic, sizeof(magic)) < 0)
-		err("Failed/2.2: %m");
-	if (write(sock, &opt, sizeof(opt)) < 0)
-		err("Failed writing abort");
-	if (write(sock, &len, sizeof(len)) < 0)
-		err("Failed writing length");
+	send_request(sock, NBD_OPT_ABORT, 0, NULL);
 }
 
-void negotiate(int sock, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts) {
-	u64 magic, size64;
+void parse_sizes(char *buf, uint64_t *size, uint16_t *flags) {
+	memcpy(size, buf, sizeof(*size));
+	*size = ntohll(*size);
+	buf += sizeof(*size);
+	memcpy(flags, buf, sizeof(*flags));
+	*flags = ntohs(*flags);
+
+	if ((*size>>12) > (uint64_t)~0UL) {
+		printf("size = %luMB", (unsigned long)(*size>>20));
+		err("Exported device is too big for me. Get 64-bit machine :-(\n");
+	} else {
+		printf("size = %luMB", (unsigned long)(*size>>20));
+	}
+	printf("\n");
+}
+
+void negotiate(int *sockp, u64 *rsize64, uint16_t *flags, char* name, uint32_t needed_flags, uint32_t client_flags, uint32_t do_opts, char *certfile, char *keyfile, char *cacertfile, char *tlshostname, bool tls) {
+	u64 magic;
 	uint16_t tmp;
 	uint16_t global_flags;
 	char buf[256] = "\0\0\0\0\0\0\0\0\0";
-	uint32_t opt;
-	uint32_t namesize;
+	int sock = *sockp;
 
 	printf("Negotiation: ");
 	readit(sock, buf, 8);
@@ -304,47 +498,168 @@ void negotiate(int sock, u64 *rsize64, uint16_t *flags, char* name, uint32_t nee
 	if (write(sock, &client_flags, sizeof(client_flags)) < 0)
 		err("Failed/2.1: %m");
 
+#if HAVE_GNUTLS && !defined(NOTLS)
+        /* TLS */
+        if (tls) {
+		int plainfd[2]; // [0] is used by the proxy, [1] is used by NBD
+		tlssession_t *s = NULL;
+		int ret;
+		uint32_t tmp32;
+		uint64_t tmp64;
+
+		send_request(sock, NBD_OPT_STARTTLS, 0, NULL);
+
+		if (read(sock, &tmp64, sizeof(tmp64)) < 0)
+			err("Could not read cliserv_magic: %m");
+		tmp64 = ntohll(tmp64);
+		if (tmp64 != NBD_OPT_REPLY_MAGIC) {
+			err("reply magic does not match");
+		}
+		if (read(sock, &tmp32, sizeof(tmp32)) < 0)
+			err("Could not read option type: %m");
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != NBD_OPT_STARTTLS)
+			err("Reply to wrong option");
+		if (read(sock, &tmp32, sizeof(tmp32)) < 0)
+			err("Could not read option reply type: %m");
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != NBD_REP_ACK) {
+			err("Option reply type != NBD_REP_ACK");
+		}
+		if (read(sock, &tmp32, sizeof(tmp32)) < 0) err(
+			"Could not read option data length: %m");
+		tmp32 = ntohl(tmp32);
+		if (tmp32 != 0) {
+			err("Option reply data length != 0");
+		}
+		s = tlssession_new(0,
+				   keyfile,
+				   certfile,
+				   cacertfile,
+				   tlshostname,
+				   !cacertfile || !tlshostname, // insecure flag
+#ifdef DODBG
+				   1, // debug
+#else
+				   0, // debug
+#endif
+				   NULL, // quitfn
+				   NULL, // erroutfn
+				   NULL // opaque
+			);
+		if (!s)
+			err("Cannot establish TLS session");
+
+		if (socketpair(AF_UNIX, SOCK_STREAM, 0, plainfd) < 0)
+			err("Cannot get socket pair");
+
+		if (set_nonblocking(plainfd[0], 0) <0 ||
+		    set_nonblocking(plainfd[1], 0) <0 ||
+		    set_nonblocking(sock, 0) <0) {
+			close(plainfd[0]);
+			close(plainfd[1]);
+			err("Cannot set socket options");
+		}
+
+		ret = fork();
+		if (ret < 0)
+			err("Could not fork");
+		else if (ret == 0) {
+			// we are the child
+			if (daemon(0, 0) < 0) {
+				/* no one will see this */
+				fprintf(stderr, "Can't detach from the terminal");
+				exit(1);
+			}
+			signal (SIGPIPE, SIG_IGN);
+			close(plainfd[1]);
+			tlssession_mainloop(sock, plainfd[0], s);
+			close(sock);
+			close(plainfd[0]);
+			exit(0);
+		}
+		close(plainfd[0]);
+		close(sock);
+		sock = plainfd[1]; /* use the decrypted FD from now on */
+		*sockp = sock;
+	}
+#else
+	if (keyfile) {
+		err("TLS requested but support not compiled in");
+	}
+#endif
+
 	if(do_opts & NBDC_DO_LIST) {
 		ask_list(sock);
 		exit(EXIT_SUCCESS);
 	}
 
-	/* Write the export name that we're after */
-	magic = htonll(opts_magic);
-	if (write(sock, &magic, sizeof(magic)) < 0)
-		err("Failed/2.2: %m");
+	send_info_request(sock, NBD_OPT_GO, 0, NULL, name);
 
-	opt = ntohl(NBD_OPT_EXPORT_NAME);
-	if (write(sock, &opt, sizeof(opt)) < 0)
-		err("Failed/2.3: %m");
-	namesize = (u32)strlen(name);
-	namesize = ntohl(namesize);
-	if (write(sock, &namesize, sizeof(namesize)) < 0)
-		err("Failed/2.4: %m");
-	if (write(sock, name, strlen(name)) < 0)
-		err("Failed/2.4: %m");
-
-	readit(sock, &size64, sizeof(size64));
-	size64 = ntohll(size64);
-
-	if ((size64>>12) > (uint64_t)~0UL) {
-		printf("size = %luMB", (unsigned long)(size64>>20));
-		err("Exported device is too big for me. Get 64-bit machine :-(\n");
-	} else
-		printf("size = %luMB", (unsigned long)(size64>>20));
-
-	readit(sock, &tmp, sizeof(tmp));
-	*flags = (uint32_t)ntohs(tmp);
-
-	if (!(global_flags & NBD_FLAG_NO_ZEROES)) {
-		readit(sock, &buf, 124);
-	}
-	printf("\n");
-
-	*rsize64 = size64;
+	struct reply *rep = NULL;
+	
+	do {
+		if(rep != NULL) free(rep);
+		rep = read_reply(sock);
+		if(rep->reply_type & NBD_REP_FLAG_ERROR) {
+			switch(rep->reply_type) {
+				case NBD_REP_ERR_UNSUP:
+					/* server doesn't support NBD_OPT_GO or NBD_OPT_INFO,
+					 * fall back to NBD_OPT_EXPORT_NAME */
+					send_request(sock, NBD_OPT_EXPORT_NAME, -1, name);
+					char b[sizeof(*flags) + sizeof(*rsize64)];
+					readit(sock, b, sizeof(b));
+					parse_sizes(b, rsize64, flags);
+					if(!(global_flags & NBD_FLAG_NO_ZEROES)) {
+						readit(sock, buf, 124);
+					}
+					free(rep);
+					return;
+				case NBD_REP_ERR_POLICY:
+					if(rep->datasize > 0) {
+						char errstr[1024];
+						snprintf(errstr, sizeof errstr, "Connection not allowed by server policy. Server said: %s", rep->data);
+						err(errstr);
+					} else {
+						err("Connection not allowed by server policy.");
+					}
+					exit(EXIT_FAILURE);
+				default:
+					if(rep->datasize > 0) {
+						char errstr[1024];
+						snprintf(errstr, sizeof errstr, "Unknown error returned by server. Server said: %s", rep->data);
+						err(errstr);
+					} else {
+						err("Unknown error returned by server.");
+					}
+					free(rep);
+					exit(EXIT_FAILURE);
+			}
+		}
+		uint16_t info_type;
+		switch(rep->reply_type) {
+			case NBD_REP_INFO:
+				memcpy(&info_type, rep->data, 2);
+				info_type = htons(info_type);
+				switch(info_type) {
+					case NBD_INFO_EXPORT:
+						parse_sizes(rep->data + 2, rsize64, flags);
+						break;
+					default:
+						// ignore these, don't need them
+						break;
+				}
+				break;
+			case NBD_REP_ACK:
+				break;
+			default:
+				err_nonfatal("Unknown reply to NBD_OPT_GO received");
+		}
+	} while(rep->reply_type != NBD_REP_ACK);
+	free(rep);
 }
 
-bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* sdp, int* b_unix, char**port) {
+bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** hostn_ptr, int* bs, int* timeout, int* persist, int* swap, int* sdp, int* b_unix, char**port, int* num_conns, char **certfile, char **keyfile, char **cacertfile, char **tlshostname) {
 	int fd = open(SYSCONFDIR "/nbdtab", O_RDONLY);
 	bool retval = false;
 	if(fd < 0) {
@@ -401,6 +716,10 @@ bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** host
 #undef MOVE_NEXT
 	// fourth field is the options field, a comma-separated field of options
 	do {
+		if(!strncmp(loc, "conns=", 6)) {
+			*num_conns = (int)strtol(loc+6, &loc, 0);
+			goto next;
+		}
 		if(!strncmp(loc, "bs=", 3)) {
 			*bs = (int)strtol(loc+3, &loc, 0);
 			goto next;
@@ -433,6 +752,22 @@ bool get_from_config(char* cfgname, char** name_ptr, char** dev_ptr, char** host
 			loc += 4;
 			goto next;
 		}
+		if(!strncmp(loc, "certfile=", 9)) {
+			*certfile = strndup(loc+9, strcspn(loc+9, ","));
+			goto next;
+		}
+		if(!strncmp(loc, "keyfile=", 8)) {
+			*keyfile = strndup(loc+8, strcspn(loc+8, ","));
+			goto next;
+		}
+		if(!strncmp(loc, "cacertfile=", 11)) {
+			*cacertfile = strndup(loc+11, strcspn(loc+11, ","));
+			goto next;
+		}
+		if(!strncmp(loc, "tlshostname=", 9)) {
+			*tlshostname = strndup(loc+9, strcspn(loc+9, ","));
+			goto next;
+		}
 		// skip unknown options, with a warning unless they start with a '_'
 		l = strcspn(loc, ",");
 		if(*loc != '_') {
@@ -482,7 +817,7 @@ void setsizes(int nbd, u64 size64, int blocksize, u32 flags) {
 				err("Ioctl/1.1c failed: %m\n");
 			}
 		}
-		fprintf(stderr, "bs=%d, sz=%llu bytes\n", blocksize, (u64)tmp_blocksize * size);
+		fprintf(stderr, "bs=%d, sz=%" PRIu64 " bytes\n", blocksize, (u64)tmp_blocksize * size);
 	}
 
 	ioctl(nbd, NBD_CLEAR_SOCK);
@@ -503,8 +838,12 @@ void set_timeout(int nbd, int timeout) {
 }
 
 void finish_sock(int sock, int nbd, int swap) {
-	if (ioctl(nbd, NBD_SET_SOCK, sock) < 0)
-		err("Ioctl NBD_SET_SOCK failed: %m\n");
+	if (ioctl(nbd, NBD_SET_SOCK, sock) < 0) {
+		if (errno == EBUSY)
+			err("Kernel doesn't support multiple connections\n");
+		else
+			err("Ioctl NBD_SET_SOCK failed: %m\n");
+	}
 
 #ifndef __ANDROID__
 	if (swap)
@@ -536,20 +875,29 @@ void usage(char* errmsg, ...) {
 		vfprintf(stderr, tmp, ap);
 		va_end(ap);
 	} else {
-		fprintf(stderr, "nbd-client version %s\n", PACKAGE_VERSION);
+		fprintf(stderr, "%s version %s\n", PROG_NAME, PACKAGE_VERSION);
 	}
+#if HAVE_NETLINK
+	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m] [-nonetlink|-L]\n");
+#else
 	fprintf(stderr, "Usage: nbd-client -name|-N name host [port] nbd_device\n\t[-block-size|-b block size] [-timeout|-t timeout] [-swap|-s] [-sdp|-S]\n\t[-persist|-p] [-nofork|-n] [-systemd-mark|-m]\n");
+#endif
 	fprintf(stderr, "Or   : nbd-client -u (with same arguments as above)\n");
 	fprintf(stderr, "Or   : nbd-client nbdX\n");
 	fprintf(stderr, "Or   : nbd-client -d nbd_device\n");
 	fprintf(stderr, "Or   : nbd-client -c nbd_device\n");
 	fprintf(stderr, "Or   : nbd-client -h|--help\n");
 	fprintf(stderr, "Or   : nbd-client -l|--list host\n");
+	fprintf(stderr, "Or   : nbd-client -V|--version\n");
+#if HAVE_GNUTLS && !defined(NOTLS)
+	fprintf(stderr, "All commands that connect to a host also take:\n\t[-F|-certfile certfile] [-K|-keyfile keyfile]\n\t[-A|-cacertfile cacertfile] [-H|-tlshostname hostname] [-x|-enable-tls]\n");
+#endif
 	fprintf(stderr, "Default value for blocksize is 1024 (recommended for ethernet)\n");
 	fprintf(stderr, "Allowed values for blocksize are 512,1024,2048,4096\n"); /* will be checked in kernel :) */
 	fprintf(stderr, "Note, that kernel 2.4.2 and older ones do not work correctly with\n");
 	fprintf(stderr, "blocksizes other than 1024 without patches\n");
 	fprintf(stderr, "Default value for port is 10809. Note that port must always be numeric\n");
+	fprintf(stderr, "Bug reports and general discussion should go to %s\n", PACKAGE_BUGREPORT);
 }
 
 void disconnect(char* device) {
@@ -566,6 +914,12 @@ void disconnect(char* device) {
 	printf("done\n");
 }
 
+#if HAVE_NETLINK
+static const char *short_opts = "-A:b:c:C:d:H:hK:LlnN:pSst:uVx";
+#else
+static const char *short_opts = "-A:b:c:C:d:H:hK:lnN:pSst:uVx";
+#endif
+
 int main(int argc, char *argv[]) {
 	char* port=NBD_DEFAULT_PORT;
 	int sock, nbd;
@@ -588,14 +942,27 @@ int main(int argc, char *argv[]) {
 	uint32_t cflags=NBD_FLAG_C_FIXED_NEWSTYLE;
 	uint32_t opts=0;
 	sigset_t block, old;
+	char *certfile = NULL;
+	char *keyfile = NULL;
+	char *cacertfile = NULL;
+	char *tlshostname = NULL;
+	bool tls = false;
 	struct sigaction sa;
+	int num_connections = 1;
+	int netlink = HAVE_NETLINK;
+	int need_disconnect = 0;
+	int *sockfds;
 	struct option long_options[] = {
 		{ "block-size", required_argument, NULL, 'b' },
 		{ "check", required_argument, NULL, 'c' },
+		{ "connections", required_argument, NULL, 'C'},
 		{ "disconnect", required_argument, NULL, 'd' },
 		{ "help", no_argument, NULL, 'h' },
 		{ "list", no_argument, NULL, 'l' },
 		{ "name", required_argument, NULL, 'N' },
+#if HAVE_NETLINK
+		{ "nonetlink", no_argument, NULL, 'L' },
+#endif
 		{ "nofork", no_argument, NULL, 'n' },
 		{ "persist", no_argument, NULL, 'p' },
 		{ "sdp", no_argument, NULL, 'S' },
@@ -603,12 +970,23 @@ int main(int argc, char *argv[]) {
 		{ "systemd-mark", no_argument, NULL, 'm' },
 		{ "timeout", required_argument, NULL, 't' },
 		{ "unix", no_argument, NULL, 'u' },
+		{ "certfile", required_argument, NULL, 'F' },
+		{ "keyfile", required_argument, NULL, 'K' },
+		{ "cacertfile", required_argument, NULL, 'A' },
+		{ "tlshostname", required_argument, NULL, 'H' },
+		{ "enable-tls", no_argument, NULL, 'x' },
+		{ "version", no_argument, NULL, 'V' },
 		{ 0, 0, 0, 0 }, 
 	};
+	int i;
 
 	logging(MY_NAME);
 
-	while((c=getopt_long_only(argc, argv, "-b:c:d:hlnN:pSst:u", long_options, NULL))>=0) {
+#if HAVE_GNUTLS && !defined(NOTLS)
+        tlssession_init();
+#endif
+
+	while((c=getopt_long_only(argc, argv, short_opts, long_options, NULL))>=0) {
 		switch(c) {
 		case 1:
 			// non-option argument
@@ -657,9 +1035,13 @@ int main(int argc, char *argv[]) {
 			break;
 		case 'c':
 			return check_conn(optarg, 1);
+		case 'C':
+			num_connections = (int)strtol(optarg, NULL, 0);
+			break;
 		case 'd':
-			disconnect(optarg);
-			exit(EXIT_SUCCESS);
+			need_disconnect = 1;
+			nbddev = strdup(optarg);
+			break;
 		case 'h':
 			usage(NULL);
 			exit(EXIT_SUCCESS);
@@ -668,6 +1050,11 @@ int main(int argc, char *argv[]) {
 			opts |= NBDC_DO_LIST;
 			nbddev="";
 			break;
+#if HAVE_NETLINK
+		case 'L':
+			netlink = 0;
+			break;
+#endif
 		case 'm':
 			argv[0][0] = '@';
 			break;
@@ -693,12 +1080,46 @@ int main(int argc, char *argv[]) {
 		case 'u':
 			b_unix = 1;
 			break;
+		case 'V':
+			printf("This is %s, from %s\n", PROG_NAME, PACKAGE_STRING);
+			return 0;
+#if HAVE_GNUTLS && !defined(NOTLS)
+		case 'x':
+			tls = true;
+			break;
+                case 'F':
+                        certfile=strdup(optarg);
+                        break;
+                case 'K':
+                        keyfile=strdup(optarg);
+                        break;
+                case 'A':
+                        cacertfile=strdup(optarg);
+                        break;
+                case 'H':
+                        tlshostname=strdup(optarg);
+                        break;
+#else
+                case 'F':
+                case 'K':
+                case 'H':
+                case 'A':
+			fprintf(stderr, "E: TLS support not compiled in\n");
+                        exit(EXIT_FAILURE);
+#endif
 		default:
 			fprintf(stderr, "E: option eaten by 42 mice\n");
 			exit(EXIT_FAILURE);
 		}
 	}
 
+	if (need_disconnect) {
+		if (netlink)
+			netlink_disconnect(nbddev);
+		else
+			disconnect(nbddev);
+		exit(EXIT_SUCCESS);
+	}
 #ifdef __ANDROID__
   if (swap)
     err("swap option unsupported on Android because mlockall is unsupported.");
@@ -706,11 +1127,11 @@ int main(int argc, char *argv[]) {
 	if(hostname) {
 		if((!name || !nbddev) && !(opts & NBDC_DO_LIST)) {
 			if(!strncmp(hostname, "nbd", 3) || !strncmp(hostname, "/dev/nbd", 8)) {
-				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &sdp, &b_unix, &port)) {
+				if(!get_from_config(hostname, &name, &nbddev, &hostname, &blocksize, &timeout, &cont, &swap, &sdp, &b_unix, &port, &num_connections, &certfile, &keyfile, &cacertfile, &hostname)) {
 					usage("no valid configuration for specified device found", hostname);
 					exit(EXIT_FAILURE);
 				}
-			} else {
+			} else if (!netlink) {
 				usage("not enough information specified, and argument didn't look like an nbd device");
 				exit(EXIT_FAILURE);
 			}
@@ -720,34 +1141,75 @@ int main(int argc, char *argv[]) {
 		exit(EXIT_FAILURE);
 	}
 
+        if (keyfile && !certfile)
+		certfile = strdup(keyfile);
+
+	if (certfile != NULL || keyfile != NULL || cacertfile != NULL || tlshostname != NULL) {
+		tls = true;
+	}
+
+        if (!tlshostname && hostname)
+                tlshostname = strdup(hostname);
+
+	if (netlink)
+		nofork = 1;
+
 	if(strlen(name)==0 && !(opts & NBDC_DO_LIST)) {
 		printf("Warning: the oldstyle protocol is no longer supported.\nThis method now uses the newstyle protocol with a default export\n");
 	}
 
-	if (b_unix)
-		sock = openunix(hostname);
-	else
-		sock = opennet(hostname, port, sdp);
-	if (sock < 0)
-		exit(EXIT_FAILURE);
+	if(!(opts & NBDC_DO_LIST) && !netlink) {
+		nbd = open(nbddev, O_RDWR);
+		if (nbd < 0)
+			err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
+	}
 
-	negotiate(sock, &size64, &flags, name, needed_flags, cflags, opts);
+	if (netlink) {
+		sockfds = malloc(sizeof(int) * num_connections);
+		if (!sockfds)
+			err("Cannot allocate the socket fd's array");
+	}
 
-	nbd = open(nbddev, O_RDWR);
-	if (nbd < 0)
-	  err("Cannot open NBD: %m\nPlease ensure the 'nbd' module is loaded.");
+	for (i = 0; i < num_connections; i++) {
+		if (b_unix)
+			sock = openunix(hostname);
+		else
+			sock = opennet(hostname, port, sdp);
+		if (sock < 0)
+			exit(EXIT_FAILURE);
 
-	setsizes(nbd, size64, blocksize, flags);
-	set_timeout(nbd, timeout);
-	finish_sock(sock, nbd, swap);
-	if (swap) {
-		/* try linux >= 2.6.36 interface first */
-		if (oom_adjust("/proc/self/oom_score_adj", "-1000")) {
-			/* fall back to linux <= 2.6.35 interface */
-			oom_adjust("/proc/self/oom_adj", "-17");
+		negotiate(&sock, &size64, &flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
+		if (netlink) {
+			sockfds[i] = sock;
+			continue;
+		}
+
+		if (i == 0) {
+			setsizes(nbd, size64, blocksize, flags);
+			set_timeout(nbd, timeout);
+		}
+		finish_sock(sock, nbd, swap);
+		if (swap) {
+			if (keyfile)
+				fprintf(stderr, "Warning: using swap and TLS is prone to deadlock\n");
+			/* try linux >= 2.6.36 interface first */
+			if (oom_adjust("/proc/self/oom_score_adj", "-1000")) {
+				/* fall back to linux <= 2.6.35 interface */
+				oom_adjust("/proc/self/oom_adj", "-17");
+			}
 		}
 	}
 
+	if (netlink) {
+		int index = -1;
+		if (nbddev) {
+			if (sscanf(nbddev, "/dev/nbd%d", &index) != 1)
+				err("Invalid nbd device target\n");
+		}
+		netlink_configure(index, sockfds, num_connections,
+				  size64, blocksize, flags, timeout);
+		return 0;
+	}
 	/* Go daemon */
 	
 #ifndef NOFORK
@@ -797,7 +1259,9 @@ int main(int argc, char *argv[]) {
 				}
 				nanosleep(&req, NULL);
 			}
-			open(nbddev, O_RDONLY);
+			if(open(nbddev, O_RDONLY) < 0) {
+				perror("could not open device for updating partition table");
+			}
 			exit(0);
 		}
 #endif
@@ -828,7 +1292,7 @@ int main(int argc, char *argv[]) {
 					nbd = open(nbddev, O_RDWR);
 					if (nbd < 0)
 						err("Cannot open NBD: %m");
-					negotiate(sock, &new_size, &new_flags, name, needed_flags, cflags, opts);
+					negotiate(&sock, &new_size, &new_flags, name, needed_flags, cflags, opts, certfile, keyfile, cacertfile, tlshostname, tls);
 					if (size64 != new_size) {
 						err("Size of the device changed. Bye");
 					}
